/*
Archive.java
This file contains the class Archive and all its functionality.

The file format of an archive is as follows:

HEADER
8 bytes:  "<410zip>"
4 bytes:  int n, number of files in archive
File 1 info:
  4 bytes:  int k1, length of name of file 1
  k1 bytes: string, name of file 1
  4 bytes:  int a1, address of compressed data for file 1
  4 bytes:  int u1, uncompressed size of file 1
File 2 info:
  4 bytes:  int k2, length of name of file 2
  k2 bytes: string, name of file 2
  4 bytes:  int a2, address of compressed data for file 2
  4 bytes:  int u2, uncompressed size of file 2
...
File n info:
  4 bytes:  int kn, length of name of file n
  kn bytes: string, name of file n
  4 bytes:  int an, address of compressed data for file n
  4 bytes:  int un, uncompressed size of file n

DATA
a1 to a2-1: compressed data for file 1
a2 to a3-1: compressed data for file 2
...
an to end of file: compressed data for file n
*/

import java.io.*;
import java.awt.*;

class Archive extends File {
	static final int maxFiles = 256; //max # of files
	String name; //name of this archive
	int n; //number of files in this archive
	String[] fileName = new String[maxFiles]; //names of files
	int[] address = new int[maxFiles]; //starting byte in archive
	int[] uncompressedSize = new int[maxFiles]; //uncompressed size of file
	int headerLength; //archive header length
	LZ lz; //the LZ trie
	static Archive archive = null; //the current archive
		 
	Archive(String fileName) {
		super(fileName);
		name = fileName;
	}
	
	//create new archive
	static void newArchive() {
		FileDialog openDialog = new FileDialog(Zip.frame,"New Archive",FileDialog.LOAD);
		openDialog.show();
		if (openDialog.getFile() == null) return; //user clicked cancel
		archive = new Archive(openDialog.getDirectory() + openDialog.getFile());
		if (archive.exists()) {
			Zip.message("Archive already exists");
			return;
		} 
		archive.n = 0; //no files
		archive.writeHeader(archive); //write out header info
		try {
			IO.closeForWrite();
		} catch (IOException e) {
			Zip.message("Cannot write archive");
			return;
		}
		archive.read(); //read it back in
	}
	
	//open existing archive	
	static public void openArchive() {
		FileDialog openDialog = new FileDialog(Zip.frame,"Open Archive",FileDialog.LOAD);
		openDialog.show();
		if (openDialog.getFile() == null) return; //user clicked cancel
		archive = new Archive(openDialog.getDirectory() + openDialog.getFile());
		if (!archive.exists()) { //archive must exist
			Zip.message("Archive does not exist");
			return;
		}
		else if (!archive.isFile()) { //must be a file, not a directory
			Zip.message("Not a valid archive");
			return;
		}
		else if (!archive.canRead()) { //must be readable
			Zip.message("Archive cannot be read");
			return;
		}
		else if (!archive.canWrite()) { //and writable
			Zip.message("Archive cannot be written");
			return;
		}
		archive.read(); //read it in
	}
	
	//read in archive
	void read() {
		int i,m;
		try {
			IO.openForRead(this);
			if (!IO.getString(8).equals("<410zip>")) { //1st 8 bytes
				Zip.message("Not a valid archive");
				return;
			}
			n = IO.getInteger(); //number of files in archive
			if (n < 0 || n > maxFiles) {
				Zip.message("Not a valid archive");
				return;
			}
			for (i = 0; i < n; i++) { //for each file in archive
				m = IO.getInteger(); //length of file name
				if (m < 0 || m > 256) {
					Zip.message("Not a valid archive");
					return;
				}
				fileName[i] = IO.getString(m); //get file name
				address[i] = IO.getInteger(); //byte address in compressed file
				uncompressedSize[i] = IO.getInteger(); //uncompressed size
			}
		} catch (IOException e) {
			Zip.message("Not a valid archive");
			return;
		}
		headerLength = IO.bytesRead; //length of header
		String fileList = ""; //assemble file info for display
		for (i = 0; i < n; i++) {
			fileList += fileName[i] + " ";
			fileList += compressedSize(i) + "/";
			fileList += uncompressedSize[i] + " ";
			double compressionRatio = Math.round(compressedSize(i)*1000.0/uncompressedSize[i])/10.0;
			fileList += compressionRatio + "%\n";
		}			
		Zip.listFiles(fileList); //display file info
		Zip.setTitle(getAbsolutePath());
		Zip.clearMessage();
		Zip.setOpenState();
	}
	
	//add a file to the archive
	void addFile() {
		FileDialog openDialog = new FileDialog(Zip.frame,"Select File to Add",FileDialog.LOAD);
		openDialog.show();
		if (openDialog.getFile() == null) return; //user clicked cancel
		File f = new File(openDialog.getDirectory() + openDialog.getFile());
		if (!f.exists()) { //file better exist
			Zip.message("File does not exist");
			return;
		}
		else if (!f.isFile()) { //and better not be a directory
			Zip.message("Not a valid file");
			return;
		}
		else if (!f.canRead()) { //and better be readable
			Zip.message("File cannot be read");
			return;
		}
		File temp = new Archive("C:\\Temp\\" + getName()); //temp file
		fileName[n] = f.getName(); //add new file info to header
		address[n] = (int)length() - headerLength;
		uncompressedSize[n++] = (int)f.length();
		lz = new LZ(); //initialize compression tables
		try {
			writeHeader(temp); //write new archive header to temp file
			IO.copy(); //copy previous compressed files to temp file
			IO.closeForRead();
			IO.openForRead(f); //encode new file and write to temp file
			lz.encode();
			IO.cleanup();
			IO.openForRead(temp); //now copy back to archive
			IO.openForWrite(this);
			IO.copy(); //should really do this with a system call
			IO.cleanup();
		} catch (IOException e) {
			Zip.message("Could not write archive: " + e.getMessage());
		}
		read(); //reread the archive
	}

	//extract files from the archive to C:\Temp\
	void extract() {
		File dir = new File("C:\\Temp");
		String dirName = dir.getAbsolutePath() + File.separator;
		if (!dir.exists()) { //check that directory exists and is writable
			Zip.message("Directory " + dirName + " does not exist");
			return;
		}
		else if (dir.isFile()) {
			Zip.message(dirName + " is not a directory");
			return;
		}
		else if (!dir.canRead()) {
			Zip.message("Cannot read " + dirName);
			return;
		}
		else if (!dir.canWrite()) {
			Zip.message("Cannot write to " + dirName);
			return;
		}
		Zip.message("Extracting files to " + dirName + " ... ");
		try {
			for (int i = 0; i < n; i++) { //for each file in archive
				IO.openForWrite(new File(dirName + fileName[i]));
				lz = new LZ(); //initialize compression tables
				lz.decode(compressedSize(i)); //decompress it
				IO.closeForWrite();
			}
			IO.closeForRead();
		} catch (IOException e) {
			Zip.message("Could not extract files: " + e.getMessage());
			return;
		}
		read(); //reread the archive
		Zip.message("Files extracted to " + dirName);
	}
	
	//get compressed size of i-th file in archive
	//address[i+1] - address[i] if not the last file
	//end of file - address[i] if the last file	
	int compressedSize(int i) {
		return ((i < n-1)?address[i+1]:(int)length()-headerLength) - address[i];
	}
	
	//write out header to archive
	void writeHeader(File archive) {
		try {
			IO.openForWrite(archive);
			IO.putString("<410zip>");
			IO.putInteger(n); //number of files
			for (int i = 0; i < n; i++) {
				IO.putInteger(fileName[i].length()); //length of filename string
				IO.putString(fileName[i]); //filename
				IO.putInteger(address[i]); //address of compressed data for this file
				IO.putInteger(uncompressedSize[i]);
			}
			IO.flush();
		} catch (IOException e) {
			Zip.message("Could not write archive");
		}
	}
	
	//close the archive	
	void close() {
		try {
			IO.closeForRead();
		} catch (IOException e) {
			Zip.message(e.getMessage());
		}
		Zip.setCloseState();
		Zip.clearFiles();
		Zip.clearMessage();
		archive = null;
	}
}
