/*
LZ.java
This file contains all code for LZ compression and decompression.
*/

import java.io.*;
import java.awt.*;
import java.util.*;

/*
The Dictionary class maintains a one-to-one correspondence
between Strings and codes.  A hashtable is used to retrieve
the code corresponding to a given String, and a Vector is used
to retrieve the String corresponding to a given code.
*/
class Dictionary {
	Hashtable hash; //Strings -> codes
	Vector vect; //codes -> Strings
	static final int maxCode = 256*256; //codes are 2 bytes
	int nextCode; //codes are assigned in order
	
	//initialize with all single-letter strings
	Dictionary() {
		hash = new Hashtable();
		vect = new Vector();
		nextCode = 0;
		char[] c = new char[1];
		for (int i = 0; i < 256; i++) {
			c[0] = (char)i;
			insert(new String(c));
		}
	}
	
	void insert(String s) {
		if (nextCode >= maxCode) return; //just return if dictionary is full
		if (contains(s)) return; //or if String already has a code
		hash.put(s,new Integer(nextCode++)); //the given String gets next code
		vect.addElement(s); //add it to both data structures
		if (vect.size() != nextCode) { //sanity check
			throw new Error("Dictionary corrupted");
		}
	}
	
	boolean contains(String s) { //check if Dictionary contains a given string
		return hash.containsKey(s);
	}
	
	boolean contains(int code) { //check if a given code has been assigned
		return code < vect.size();
	}

	int getCode(String s) { //String -> code
		return ((Integer)hash.get(s)).intValue();
	}

	String getString(int code) { //code -> String
		return (String)vect.elementAt(code);
	}
}

/*
This class contains the code for encoding and decoding.
*/
class LZ {
	StringBuffer buf = new StringBuffer();
	int lastCharRead; //latest character read
	String lastMatchedString; //longest prefix of input stream in dictionary
	String lastUnmatchedString; //that + following character
	String onDeck; //next String to be inserted in the dictionary
	Dictionary dict; //the dictionary
	
	LZ() { //initialize for encoding and decoding
		dict = new Dictionary();
		onDeck = null;
		lastUnmatchedString = null;
	}
	
	//encode the input stream
	void encode() throws IOException {
		lastCharRead = IO.get();
		while (lastCharRead != -1) {
			int code = longestMatch(); //longest prefix of input in dictionary
			IO.put(code/256); //write its 2-byte code
			IO.put(code%256); //to output stream
			if (onDeck != null) { //anything to add to dictionary?
				dict.insert(onDeck);
			}
			onDeck = lastUnmatchedString;
		}
	}
	
	//find the longest prefix of the input stream
	//contained in the dictionary
	int longestMatch() {
		buf.setLength(0); //start with String of length 1 containing
		buf.append((char)lastCharRead); //last character read
		lastUnmatchedString = new String(buf);
		while (dict.contains(lastUnmatchedString)) { //in dictionary?
			lastMatchedString = lastUnmatchedString; //yes--call it matched
			lastCharRead = IO.get(); //get following character
			if (lastCharRead == -1) {
				lastUnmatchedString = null; //end of file--nothing unmatched
				break;
			}
			buf.append((char)lastCharRead); //append last char and try again
			lastUnmatchedString = new String(buf);
		}
		return dict.getCode(lastMatchedString); //code of longest matched string
	}
	
	void decode(int m) throws IOException {
		String previous = null;
		while (m > 0) {
			int hi = IO.get();
			int lo = IO.get();
			m = m - 2;
			if (hi == -1 || lo == -1) throw new IOException("Unexpected end of file");
			int code = hi*256 + lo;
			if (!dict.contains(code)) throw new IOException("Invalid archive");
			String s = dict.getString(code);
			IO.putString(s);
			if (previous != null) dict.insert(previous + s.charAt(0));
			previous = s;
		}
	}
}